package Database::Async::Test;

use strict;
use warnings;

our $VERSION = '0.016'; # VERSION

=head1 NAME

Database::Async::Test - utilities for testing with L<Database::Async>.

=head1 DESCRIPTION

=cut

1;

=head1 AUTHOR

Tom Molesworth C<< <TEAM@cpan.org> >>

=head1 LICENSE

Copyright Tom Molesworth 2011-2021. Licensed under the same terms as Perl itself.

